import "@dxos/node-std/globals";

// inject-globals:@inject-globals
import {
  global,
  Buffer,
  process
} from "@dxos/node-std/inject-globals";

// packages/core/halo/credentials/src/invitations.ts
var defaultSecretProvider = async () => Buffer.from("0000");
var defaultSecretValidator = async (invitation, secret) => true;

// packages/core/halo/credentials/src/passcode.ts
var generatePasscode = (length = 4) => {
  let passcode = "";
  for (let i = 0; i < length; i++) {
    passcode += `${Math.floor(Math.random() * 10)}`;
  }
  return passcode;
};

// packages/core/halo/credentials/src/credentials/signing.ts
import stableStringify from "json-stable-stringify";
import { PublicKey } from "@dxos/keys";
import { Timeframe } from "@dxos/timeframe";
import { arrayToBuffer } from "@dxos/util";
var getCredentialProofPayload = (credential) => {
  const copy = {
    ...credential,
    proof: {
      ...credential.proof,
      value: new Uint8Array(),
      chain: void 0
    }
  };
  delete copy.id;
  return Buffer.from(canonicalStringify(copy));
};
var canonicalStringify = (obj) => stableStringify(obj, {
  /* The point of signing and verifying is not that the internal, private state of the objects be
  * identical, but that the public contents can be verified not to have been altered. For that reason,
  * really private fields (indicated by '__') are not included in the signature.
  * This gives a mechanism for attaching other attributes to an object without breaking the signature.
  * We also skip @type.
  */
  // TODO(dmaretskyi): Should we actually skip the @type field?
  replacer: function(key, value) {
    if (key.toString().startsWith("__") || key.toString() === "@type") {
      return void 0;
    }
    if (value === null) {
      return void 0;
    }
    const original = this[key];
    if (value) {
      if (PublicKey.isPublicKey(value)) {
        return value.toHex();
      }
      if (Buffer.isBuffer(value)) {
        return value.toString("hex");
      }
      if (value instanceof Uint8Array) {
        return arrayToBuffer(value).toString("hex");
      }
      if (value.data && value.type === "Buffer") {
        return Buffer.from(value).toString("hex");
      }
      if (original instanceof Timeframe) {
        return original.frames().reduce((frames, [key2, seq]) => {
          frames[truncateKey(key2)] = seq;
          return frames;
        }, {});
      }
    }
    return value;
  }
});
var truncateKey = (key) => {
  const str = key.toHex();
  return `${str.substring(0, 4)}...${str.substring(str.length - 4)}`;
};

// packages/core/halo/credentials/src/presentations/signing.ts
var getPresentationProofPayload = (credentials, proof) => {
  const copy = {
    credentials,
    proof: {
      ...proof,
      value: new Uint8Array(),
      chain: void 0
    }
  };
  return Buffer.from(canonicalStringify(copy));
};

// packages/core/halo/credentials/src/credentials/assertions.ts
var getCredentialAssertion = (credential) => credential.subject.assertion;
var isValidAuthorizedDeviceCredential = (credential, identityKey, deviceKey) => {
  const assertion = getCredentialAssertion(credential);
  return credential.subject.id.equals(deviceKey) && credential.issuer.equals(identityKey) && assertion["@type"] === "dxos.halo.credentials.AuthorizedDevice" && assertion.identityKey.equals(identityKey) && assertion.deviceKey.equals(deviceKey);
};
var checkCredentialType = (credential, type) => credential.subject.assertion["@type"] === type;
var credentialTypeFilter = (type) => (credential) => checkCredentialType(credential, type);

// packages/core/halo/credentials/src/credentials/credential-factory.ts
import { subtleCrypto } from "@dxos/crypto";
import { invariant } from "@dxos/invariant";
import { PublicKey as PublicKey2 } from "@dxos/keys";

// packages/core/halo/credentials/src/credentials/verifier.ts
import { verifySignature } from "@dxos/crypto";
var SIGNATURE_TYPE_ED25519 = "ED25519Signature";
var verifyCredential = async (credential) => {
  if (!credential.issuer.equals(credential.proof.signer)) {
    if (!credential.proof.chain) {
      return {
        kind: "fail",
        errors: [
          "Delegated credential is missing credential chain."
        ]
      };
    }
    const result2 = await verifyChain(credential.proof.chain, credential.issuer, credential.proof.signer);
    if (result2.kind === "fail") {
      return result2;
    }
  }
  const result = await verifyCredentialSignature(credential);
  if (result.kind === "fail") {
    return result;
  }
  return {
    kind: "pass"
  };
};
var verifyCredentialSignature = async (credential) => {
  if (credential.proof.type !== SIGNATURE_TYPE_ED25519) {
    return {
      kind: "fail",
      errors: [
        `Invalid signature type: ${credential.proof.type}`
      ]
    };
  }
  const signData = getCredentialProofPayload(credential);
  if (!await verifySignature(credential.proof.signer, signData, credential.proof.value)) {
    return {
      kind: "fail",
      errors: [
        "Invalid signature"
      ]
    };
  }
  return {
    kind: "pass"
  };
};
var verifyChain = async (chain, authority, subject) => {
  const result = await verifyCredential(chain.credential);
  if (result.kind === "fail") {
    return result;
  }
  if (!isValidAuthorizedDeviceCredential(chain.credential, authority, subject)) {
    return {
      kind: "fail",
      errors: [
        `Invalid credential chain: invalid assertion for key: ${subject}`
      ]
    };
  }
  return {
    kind: "pass"
  };
};

// packages/core/halo/credentials/src/credentials/credential-factory.ts
var __dxlog_file = "/home/runner/work/dxos/dxos/packages/core/halo/credentials/src/credentials/credential-factory.ts";
var createCredential = async ({ signer, issuer, subject, assertion, signingKey, chain, nonce }) => {
  invariant(assertion["@type"], "Invalid assertion.", {
    F: __dxlog_file,
    L: 45,
    S: void 0,
    A: [
      "assertion['@type']",
      "'Invalid assertion.'"
    ]
  });
  invariant(!!signingKey === !!chain, "Chain must be provided if and only if the signing key differs from the issuer.", {
    F: __dxlog_file,
    L: 46,
    S: void 0,
    A: [
      "!!signingKey === !!chain",
      "'Chain must be provided if and only if the signing key differs from the issuer.'"
    ]
  });
  if (chain) {
    const result = await verifyChain(chain, issuer, signingKey);
    invariant(result.kind === "pass", "Invalid chain.", {
      F: __dxlog_file,
      L: 49,
      S: void 0,
      A: [
        "result.kind === 'pass'",
        "'Invalid chain.'"
      ]
    });
  }
  const credential = {
    issuer,
    issuanceDate: /* @__PURE__ */ new Date(),
    subject: {
      id: subject,
      assertion
    },
    proof: {
      type: SIGNATURE_TYPE_ED25519,
      creationDate: /* @__PURE__ */ new Date(),
      signer: signingKey ?? issuer,
      value: new Uint8Array(),
      nonce
    }
  };
  const signedPayload = getCredentialProofPayload(credential);
  credential.proof.value = await signer.sign(signingKey ?? issuer, signedPayload);
  if (chain) {
    credential.proof.chain = chain;
  }
  credential.id = PublicKey2.from(await subtleCrypto.digest("SHA-256", signedPayload));
  return credential;
};
var createCredentialMessage = (credential) => {
  return {
    "@type": "dxos.echo.feed.CredentialsMessage",
    credential
  };
};
var createCredentialSignerWithKey = (signer, issuer) => ({
  getIssuer: () => issuer,
  createCredential: ({ subject, assertion, nonce }) => createCredential({
    signer,
    issuer,
    subject,
    assertion,
    nonce
  })
});
var createCredentialSignerWithChain = (signer, chain, signingKey) => ({
  getIssuer: () => chain.credential.issuer,
  createCredential: ({ subject, assertion, nonce }) => createCredential({
    signer,
    issuer: chain.credential.issuer,
    signingKey,
    chain,
    subject,
    assertion,
    nonce
  })
});

// packages/core/halo/credentials/src/credentials/credential-generator.ts
import { AdmittedFeed, SpaceMember } from "@dxos/protocols/proto/dxos/halo/credentials";
import { Timeframe as Timeframe2 } from "@dxos/timeframe";
var CredentialGenerator = class {
  constructor(_signer, _identityKey, _deviceKey) {
    this._signer = _signer;
    this._identityKey = _identityKey;
    this._deviceKey = _deviceKey;
  }
  /**
  * Create genesis messages for new Space.
  */
  async createSpaceGenesis(spaceKey, controlKey, creatorProfile) {
    return [
      await createCredential({
        signer: this._signer,
        issuer: spaceKey,
        subject: spaceKey,
        assertion: {
          "@type": "dxos.halo.credentials.SpaceGenesis",
          spaceKey
        }
      }),
      await createCredential({
        signer: this._signer,
        issuer: spaceKey,
        subject: this._identityKey,
        assertion: {
          "@type": "dxos.halo.credentials.SpaceMember",
          spaceKey,
          role: SpaceMember.Role.ADMIN,
          profile: creatorProfile,
          genesisFeedKey: controlKey
        }
      }),
      await this.createFeedAdmission(spaceKey, controlKey, AdmittedFeed.Designation.CONTROL)
    ];
  }
  /**
  * Create invitation.
  * Admit identity and control and data feeds.
  */
  // TODO(burdon): Reconcile with above (esp. Signer).
  async createMemberInvitation(spaceKey, identityKey, deviceKey, controlKey, dataKey, genesisFeedKey) {
    return [
      await createCredential({
        signer: this._signer,
        issuer: this._identityKey,
        subject: identityKey,
        assertion: {
          "@type": "dxos.halo.credentials.SpaceMember",
          spaceKey,
          role: SpaceMember.Role.MEMBER,
          genesisFeedKey
        }
      }),
      await this.createFeedAdmission(spaceKey, controlKey, AdmittedFeed.Designation.CONTROL),
      await this.createFeedAdmission(spaceKey, dataKey, AdmittedFeed.Designation.DATA)
    ];
  }
  /**
  * Add device to space.
  */
  // TODO(burdon): Reconcile with below.
  async createDeviceAuthorization(deviceKey) {
    return createCredential({
      signer: this._signer,
      issuer: this._identityKey,
      subject: deviceKey,
      assertion: {
        "@type": "dxos.halo.credentials.AuthorizedDevice",
        identityKey: this._identityKey,
        deviceKey
      }
    });
  }
  /**
  * Add device metadata.
  */
  async createDeviceProfile(profile) {
    return createCredential({
      signer: this._signer,
      issuer: this._identityKey,
      subject: this._deviceKey,
      assertion: {
        "@type": "dxos.halo.credentials.DeviceProfile",
        profile
      }
    });
  }
  /**
  * Add feed to space.
  */
  async createFeedAdmission(spaceKey, feedKey, designation) {
    return createCredential({
      signer: this._signer,
      issuer: this._identityKey,
      subject: feedKey,
      assertion: {
        "@type": "dxos.halo.credentials.AdmittedFeed",
        spaceKey,
        identityKey: this._identityKey,
        deviceKey: this._deviceKey,
        designation
      }
    });
  }
  async createProfileCredential(profile) {
    return createCredential({
      signer: this._signer,
      issuer: this._identityKey,
      subject: this._identityKey,
      assertion: {
        "@type": "dxos.halo.credentials.IdentityProfile",
        profile
      }
    });
  }
  async createEpochCredential(spaceKey) {
    return createCredential({
      signer: this._signer,
      issuer: this._identityKey,
      subject: spaceKey,
      assertion: {
        "@type": "dxos.halo.credentials.Epoch",
        number: 0,
        timeframe: new Timeframe2()
      }
    });
  }
};
var createDeviceAuthorization = async (signer, identityKey, deviceKey) => {
  const credentials = await Promise.all([
    await signer.createCredential({
      subject: deviceKey,
      assertion: {
        "@type": "dxos.halo.credentials.AuthorizedDevice",
        identityKey,
        deviceKey
      }
    })
  ]);
  return credentials.map((credential) => ({
    "@type": "dxos.echo.feed.CredentialsMessage",
    credential
  }));
};
var createAdmissionCredentials = async (signer, identityKey, spaceKey, genesisFeedKey, profile, invitationCredentialId) => {
  const credentials = await Promise.all([
    await signer.createCredential({
      subject: identityKey,
      assertion: {
        "@type": "dxos.halo.credentials.SpaceMember",
        spaceKey,
        role: SpaceMember.Role.ADMIN,
        profile,
        genesisFeedKey,
        invitationCredentialId
      }
    })
  ]);
  return credentials.map((credential) => ({
    credential: {
      credential
    }
  }));
};
var createDelegatedSpaceInvitationCredential = async (signer, subject, invitation) => {
  const credential = await signer.createCredential({
    subject,
    assertion: {
      "@type": "dxos.halo.invitations.DelegateSpaceInvitation",
      invitationId: invitation.invitationId,
      authMethod: invitation.authMethod,
      swarmKey: invitation.swarmKey,
      role: invitation.role,
      guestKey: invitation.guestKey,
      expiresOn: invitation.expiresOn,
      multiUse: invitation.multiUse
    }
  });
  return {
    credential: {
      credential
    }
  };
};

// packages/core/halo/credentials/src/presentations/presentation.ts
var signPresentation = async ({ presentation, signer, signerKey, chain, nonce }) => {
  const proof = {
    type: SIGNATURE_TYPE_ED25519,
    value: new Uint8Array(),
    creationDate: /* @__PURE__ */ new Date(),
    signer: signerKey,
    nonce
  };
  const signedPayload = getPresentationProofPayload(presentation.credentials ?? [], proof);
  proof.value = await signer.sign(signerKey, signedPayload);
  if (chain) {
    proof.chain = chain;
  }
  return {
    credentials: presentation.credentials,
    proofs: [
      ...presentation.proofs ?? [],
      proof
    ]
  };
};

// packages/core/halo/credentials/src/presentations/verifier.ts
import { verifySignature as verifySignature2 } from "@dxos/crypto";
var verifyPresentation = async (presentation) => {
  const errors = [];
  const credentialsVerifications = await Promise.all(presentation.credentials?.map((credential) => verifyCredential(credential)) ?? []);
  for (const verification of credentialsVerifications) {
    if (verification.kind === "fail") {
      errors.push(...verification.errors);
    }
  }
  const proofVerification = await Promise.all(presentation.proofs?.map(async (proof) => {
    const chainVerification = await verifyPresentationChain(presentation, proof);
    if (chainVerification.kind === "fail") {
      return chainVerification;
    }
    const signatureVerification = await verifyPresentationSignature(presentation, proof);
    if (signatureVerification.kind === "fail") {
      return signatureVerification;
    }
    return {
      kind: "pass"
    };
  }) ?? []);
  for (const verification of proofVerification) {
    if (verification.kind === "fail") {
      errors.push(...verification.errors);
    }
  }
  if (errors.length === 0) {
    return {
      kind: "pass"
    };
  }
  {
    return {
      kind: "fail",
      errors
    };
  }
};
var verifyPresentationChain = async (presentation, proof) => {
  for (const credential of presentation.credentials ?? []) {
    if (!credential.issuer.equals(proof.signer)) {
      if (!proof.chain) {
        return {
          kind: "fail",
          errors: [
            "Delegated credential is missing credential chain."
          ]
        };
      }
      const chainVerification = await verifyChain(proof.chain, credential.issuer, proof.signer);
      if (chainVerification.kind === "fail") {
        return chainVerification;
      }
    }
  }
  return {
    kind: "pass"
  };
};
var verifyPresentationSignature = async (presentation, proof) => {
  if (proof.type !== SIGNATURE_TYPE_ED25519) {
    return {
      kind: "fail",
      errors: [
        `Invalid signature type: ${proof.type}`
      ]
    };
  }
  const signData = getPresentationProofPayload(presentation.credentials ?? [], proof);
  if (!await verifySignature2(proof.signer, signData, proof.value)) {
    return {
      kind: "fail",
      errors: [
        "Invalid signature"
      ]
    };
  }
  return {
    kind: "pass"
  };
};

// packages/core/halo/credentials/src/state-machine/space-state-machine.ts
import { runInContextAsync } from "@dxos/async";
import { Context } from "@dxos/context";
import { PublicKey as PublicKey6 } from "@dxos/keys";
import { log as log2 } from "@dxos/log";
import { SpaceMember as SpaceMember2 } from "@dxos/protocols/proto/dxos/halo/credentials";
import { Callback as Callback4, ComplexMap as ComplexMap4, ComplexSet as ComplexSet2 } from "@dxos/util";

// packages/core/halo/credentials/src/state-machine/feed-state-machine.ts
import { invariant as invariant2 } from "@dxos/invariant";
import { PublicKey as PublicKey3 } from "@dxos/keys";
import { Callback, ComplexMap } from "@dxos/util";
var __dxlog_file2 = "/home/runner/work/dxos/dxos/packages/core/halo/credentials/src/state-machine/feed-state-machine.ts";
var FeedStateMachine = class {
  constructor(_spaceKey) {
    this._spaceKey = _spaceKey;
    this._feeds = new ComplexMap(PublicKey3.hash);
    this.onFeedAdmitted = new Callback();
  }
  get feeds() {
    return this._feeds;
  }
  /**
  * Processes the AdmittedFeed credential.
  * Assumes the credential is already pre-verified
  * and the issuer has been authorized to issue credentials of this type.
  * @param fromFeed Key of the feed where this credential is recorded.
  */
  async process(credential, fromFeed) {
    const assertion = getCredentialAssertion(credential);
    invariant2(assertion["@type"] === "dxos.halo.credentials.AdmittedFeed", void 0, {
      F: __dxlog_file2,
      L: 47,
      S: this,
      A: [
        "assertion['@type'] === 'dxos.halo.credentials.AdmittedFeed'",
        ""
      ]
    });
    invariant2(assertion.spaceKey.equals(this._spaceKey), void 0, {
      F: __dxlog_file2,
      L: 48,
      S: this,
      A: [
        "assertion.spaceKey.equals(this._spaceKey)",
        ""
      ]
    });
    const info = {
      key: credential.subject.id,
      credential,
      assertion,
      parent: fromFeed
    };
    this._feeds.set(credential.subject.id, info);
    await this.onFeedAdmitted.callIfSet(info);
  }
};

// packages/core/halo/credentials/src/state-machine/invitation-state-machine.ts
import { PublicKey as PublicKey4 } from "@dxos/keys";
import { Callback as Callback2, ComplexMap as ComplexMap2, ComplexSet } from "@dxos/util";
var InvitationStateMachine = class {
  constructor() {
    this._invitations = new ComplexMap2(PublicKey4.hash);
    this._redeemedInvitationCredentialIds = new ComplexSet(PublicKey4.hash);
    this._cancelledInvitationCredentialIds = new ComplexSet(PublicKey4.hash);
    this.onDelegatedInvitation = new Callback2();
    this.onDelegatedInvitationRemoved = new Callback2();
  }
  get invitations() {
    return this._invitations;
  }
  async process(credential) {
    const credentialId = credential.id;
    if (credentialId == null) {
      return;
    }
    const assertion = getCredentialAssertion(credential);
    switch (assertion["@type"]) {
      case "dxos.halo.invitations.CancelDelegatedInvitation": {
        this._cancelledInvitationCredentialIds.add(assertion.credentialId);
        const existingInvitation = this._invitations.get(assertion.credentialId);
        if (existingInvitation != null) {
          this._invitations.delete(assertion.credentialId);
          await this.onDelegatedInvitationRemoved.callIfSet({
            credentialId: assertion.credentialId,
            invitation: existingInvitation
          });
        }
        break;
      }
      case "dxos.halo.invitations.DelegateSpaceInvitation": {
        if (credential.id) {
          const isExpired = assertion.expiresOn && assertion.expiresOn.getTime() < Date.now();
          const wasUsed = this._redeemedInvitationCredentialIds.has(credential.id) && !assertion.multiUse;
          const wasCancelled = this._cancelledInvitationCredentialIds.has(credential.id);
          if (isExpired || wasCancelled || wasUsed) {
            return;
          }
          const invitation = {
            ...assertion
          };
          this._invitations.set(credential.id, invitation);
          await this.onDelegatedInvitation.callIfSet({
            credentialId: credential.id,
            invitation
          });
        }
        break;
      }
      case "dxos.halo.credentials.SpaceMember": {
        if (assertion.invitationCredentialId != null) {
          this._redeemedInvitationCredentialIds.add(assertion.invitationCredentialId);
          const existingInvitation = this._invitations.get(assertion.invitationCredentialId);
          if (existingInvitation != null && !existingInvitation.multiUse) {
            this._invitations.delete(assertion.invitationCredentialId);
            await this.onDelegatedInvitationRemoved.callIfSet({
              credentialId: assertion.invitationCredentialId,
              invitation: existingInvitation
            });
          }
        }
        break;
      }
    }
  }
};

// packages/core/halo/credentials/src/state-machine/member-state-machine.ts
import { invariant as invariant3 } from "@dxos/invariant";
import { PublicKey as PublicKey5 } from "@dxos/keys";
import { log } from "@dxos/log";
import { Callback as Callback3, ComplexMap as ComplexMap3 } from "@dxos/util";
var __dxlog_file3 = "/home/runner/work/dxos/dxos/packages/core/halo/credentials/src/state-machine/member-state-machine.ts";
var MemberStateMachine = class {
  constructor(_spaceKey) {
    this._spaceKey = _spaceKey;
    this._members = new ComplexMap3(PublicKey5.hash);
    this.onMemberAdmitted = new Callback3();
  }
  get creator() {
    return this._creator;
  }
  get members() {
    return this._members;
  }
  getRole(member) {
    return this._members.get(member)?.assertion.role;
  }
  /**
  * Processes the SpaceMember credential.
  * Assumes the credential is already pre-verified and the issuer has been authorized to issue credentials of this type.
  */
  async process(credential) {
    const assertion = getCredentialAssertion(credential);
    switch (assertion["@type"]) {
      case "dxos.halo.credentials.SpaceMember": {
        invariant3(assertion.spaceKey.equals(this._spaceKey), void 0, {
          F: __dxlog_file3,
          L: 58,
          S: this,
          A: [
            "assertion.spaceKey.equals(this._spaceKey)",
            ""
          ]
        });
        if (this._members.has(credential.subject.id)) {
          return;
        }
        const info = {
          key: credential.subject.id,
          credential,
          assertion,
          removed: false,
          profile: assertion.profile
        };
        if (!this._creator) {
          this._creator = info;
        }
        this._members.set(credential.subject.id, info);
        log("member added", {
          member: credential.subject.id,
          space: this._spaceKey,
          role: assertion.role,
          profile: assertion.profile
        }, {
          F: __dxlog_file3,
          L: 79,
          S: this,
          C: (f, a) => f(...a)
        });
        await this.onMemberAdmitted.callIfSet(info);
        break;
      }
      case "dxos.halo.credentials.MemberProfile": {
        const member = this._members.get(credential.subject.id);
        if (!member) {
          log.warn("Member not found", {
            id: credential.subject.id
          }, {
            F: __dxlog_file3,
            L: 93,
            S: this,
            C: (f, a) => f(...a)
          });
        } else {
          member.profile = assertion.profile;
        }
        break;
      }
      default:
        throw new Error("Invalid assertion type");
    }
  }
  async onRevoked(revoked, revocation) {
    invariant3(revoked.id, void 0, {
      F: __dxlog_file3,
      L: 106,
      S: this,
      A: [
        "revoked.id",
        ""
      ]
    });
    for (const member of this._members.values()) {
      if (member.credential.id?.equals(revoked.id)) {
        member.removed = true;
        break;
      }
    }
  }
};

// packages/core/halo/credentials/src/state-machine/space-state-machine.ts
var __dxlog_file4 = "/home/runner/work/dxos/dxos/packages/core/halo/credentials/src/state-machine/space-state-machine.ts";
var REVOCABLE_CREDENTIALS = [
  "dxos.halo.credentials.SpaceMember"
];
var SpaceStateMachine = class {
  constructor(_spaceKey) {
    this._spaceKey = _spaceKey;
    this._members = new MemberStateMachine(this._spaceKey);
    this._feeds = new FeedStateMachine(this._spaceKey);
    this._invitations = new InvitationStateMachine();
    this._credentials = [];
    this._credentialsById = new ComplexMap4(PublicKey6.hash);
    this._processedCredentials = new ComplexSet2(PublicKey6.hash);
    this._credentialProcessors = [];
    this.onCredentialProcessed = new Callback4();
    this.onMemberAdmitted = this._members.onMemberAdmitted;
    this.onFeedAdmitted = this._feeds.onFeedAdmitted;
    this.onDelegatedInvitation = this._invitations.onDelegatedInvitation;
    this.onDelegatedInvitationRemoved = this._invitations.onDelegatedInvitationRemoved;
  }
  get creator() {
    return this._members.creator;
  }
  get members() {
    return this._members.members;
  }
  get feeds() {
    return this._feeds.feeds;
  }
  get credentials() {
    return this._credentials.map((entry) => entry.credential);
  }
  get credentialEntries() {
    return this._credentials;
  }
  get genesisCredential() {
    return this._genesisCredential;
  }
  get invitations() {
    return this._invitations.invitations;
  }
  async addCredentialProcessor(processor) {
    if (this._credentialProcessors.find((p) => p.processor === processor)) {
      throw new Error("Credential processor already added.");
    }
    const consumer = new CredentialConsumer(processor, async () => {
      for (const credential of this.credentials) {
        await consumer._process(credential);
      }
      consumer._isReadyForLiveCredentials = true;
    }, async () => {
      this._credentialProcessors = this._credentialProcessors.filter((p) => p !== consumer);
    });
    this._credentialProcessors.push(consumer);
    await consumer.open();
  }
  async removeCredentialProcessor(processor) {
    const consumer = this._credentialProcessors.find((p) => p.processor === processor);
    await consumer?.close();
  }
  getCredentialsOfType(type) {
    return this.credentials.filter((credential) => getCredentialAssertion(credential)["@type"] === type);
  }
  /**
  * @param credential Message to process.
  * @param fromFeed Key of the feed where this credential is recorded.
  */
  async process(credential, { sourceFeed, skipVerification }) {
    if (credential.id) {
      if (this._processedCredentials.has(credential.id)) {
        return true;
      }
      this._processedCredentials.add(credential.id);
    }
    if (!skipVerification) {
      const result = await verifyCredential(credential);
      if (result.kind !== "pass") {
        log2.warn(`Invalid credential: ${result.errors.join(", ")}`, void 0, {
          F: __dxlog_file4,
          L: 149,
          S: this,
          C: (f, a) => f(...a)
        });
        return false;
      }
    }
    const assertion = getCredentialAssertion(credential);
    switch (assertion["@type"]) {
      case "dxos.halo.credentials.Revocation": {
        if (!this._canRevokeCredentials(credential.issuer)) {
          log2.warn(`Space member is not authorized to invite new members: ${credential.issuer}`, void 0, {
            F: __dxlog_file4,
            L: 158,
            S: this,
            C: (f, a) => f(...a)
          });
          return false;
        }
        const toBeRevoked = this._credentialsById.get(credential.subject.id);
        if (!toBeRevoked) {
          log2.warn(`Credential to revoke not found: ${credential.subject.id}`, void 0, {
            F: __dxlog_file4,
            L: 163,
            S: this,
            C: (f, a) => f(...a)
          });
          return false;
        }
        if (!this._credentialCanBeRevoked(toBeRevoked)) {
          log2.warn(`Credential cannot be revoked: ${credential.subject.id}`, void 0, {
            F: __dxlog_file4,
            L: 167,
            S: this,
            C: (f, a) => f(...a)
          });
        }
        toBeRevoked.revoked = true;
        await this._members.onRevoked(toBeRevoked.credential, credential);
        break;
      }
      case "dxos.halo.credentials.SpaceGenesis": {
        if (this._genesisCredential) {
          log2.warn("Space already has a genesis credential.", void 0, {
            F: __dxlog_file4,
            L: 176,
            S: this,
            C: (f, a) => f(...a)
          });
          return false;
        }
        if (!credential.issuer.equals(this._spaceKey)) {
          log2.warn("Space genesis credential must be issued by space.", void 0, {
            F: __dxlog_file4,
            L: 180,
            S: this,
            C: (f, a) => f(...a)
          });
          return false;
        }
        if (!credential.subject.id.equals(this._spaceKey)) {
          log2.warn("Space genesis credential must be issued to space.", void 0, {
            F: __dxlog_file4,
            L: 184,
            S: this,
            C: (f, a) => f(...a)
          });
          return false;
        }
        this._genesisCredential = credential;
        break;
      }
      case "dxos.halo.credentials.SpaceMember": {
        if (!assertion.spaceKey.equals(this._spaceKey)) {
          break;
        }
        if (!this._genesisCredential) {
          log2.warn("Space must have a genesis credential before adding members.", void 0, {
            F: __dxlog_file4,
            L: 198,
            S: this,
            C: (f, a) => f(...a)
          });
          return false;
        }
        if (!this._canInviteNewMembers(credential.issuer)) {
          log2.warn(`Space member is not authorized to invite new members: ${credential.issuer}`, void 0, {
            F: __dxlog_file4,
            L: 202,
            S: this,
            C: (f, a) => f(...a)
          });
          return false;
        }
        await this._members.process(credential);
        await this._invitations.process(credential);
        break;
      }
      case "dxos.halo.credentials.MemberProfile": {
        if (!this._genesisCredential) {
          log2.warn("Space must have a genesis credential before adding members.", void 0, {
            F: __dxlog_file4,
            L: 213,
            S: this,
            C: (f, a) => f(...a)
          });
          return false;
        }
        await this._members.process(credential);
        break;
      }
      case "dxos.halo.credentials.AdmittedFeed": {
        if (!this._genesisCredential) {
          log2.warn("Space must have a genesis credential before admitting feeds.", void 0, {
            F: __dxlog_file4,
            L: 223,
            S: this,
            C: (f, a) => f(...a)
          });
          return false;
        }
        if (!this._canAdmitFeeds(credential.issuer)) {
          log2.warn(`Space member is not authorized to admit feeds: ${credential.issuer}`, void 0, {
            F: __dxlog_file4,
            L: 227,
            S: this,
            C: (f, a) => f(...a)
          });
          return false;
        }
        await this._feeds.process(credential, sourceFeed);
        break;
      }
      case "dxos.halo.invitations.CancelDelegatedInvitation":
      case "dxos.halo.invitations.DelegateSpaceInvitation": {
        if (!this._canInviteNewMembers(credential.issuer)) {
          log2.warn(`Invalid invitation, space member is not authorized to invite new members: ${credential.issuer}`, void 0, {
            F: __dxlog_file4,
            L: 238,
            S: this,
            C: (f, a) => f(...a)
          });
          return false;
        }
        await this._invitations.process(credential);
        break;
      }
    }
    const newEntry = {
      credential,
      sourceFeed,
      revoked: false
    };
    this._credentials.push(newEntry);
    if (credential.id) {
      this._credentialsById.set(credential.id, newEntry);
    }
    for (const processor of this._credentialProcessors) {
      if (processor._isReadyForLiveCredentials) {
        await processor._process(credential);
      }
    }
    await this.onCredentialProcessed.callIfSet(credential);
    return true;
  }
  _canInviteNewMembers(key) {
    return key.equals(this._spaceKey) || this._members.getRole(key) === SpaceMember2.Role.ADMIN;
  }
  _canAdmitFeeds(key) {
    const role = this._members.getRole(key);
    return role === SpaceMember2.Role.MEMBER || role === SpaceMember2.Role.ADMIN;
  }
  _canRevokeCredentials(key) {
    return key.equals(this._spaceKey) || this._members.getRole(key) === SpaceMember2.Role.ADMIN;
  }
  _credentialCanBeRevoked(entry) {
    return REVOCABLE_CREDENTIALS.includes(entry.credential.subject.assertion["@type"]);
  }
};
var CredentialConsumer = class {
  constructor(processor, _onOpen, _onClose) {
    this.processor = processor;
    this._onOpen = _onOpen;
    this._onClose = _onClose;
    this._ctx = new Context();
    this._isReadyForLiveCredentials = false;
  }
  /**
  * @internal
  */
  async _process(credential) {
    await runInContextAsync(this._ctx, async () => {
      await this.processor.processCredential(credential);
    });
  }
  async open() {
    if (this._ctx.disposed) {
      throw new Error("CredentialProcessor is disposed");
    }
    await this._onOpen();
  }
  async close() {
    await this._ctx.dispose();
    await this._onClose();
  }
};

// packages/core/halo/credentials/src/processor/device-state-machine.ts
import { Trigger } from "@dxos/async";
import { invariant as invariant4 } from "@dxos/invariant";
import { PublicKey as PublicKey7 } from "@dxos/keys";
import { log as log3 } from "@dxos/log";
import { ComplexMap as ComplexMap5 } from "@dxos/util";
var __dxlog_file5 = "/home/runner/work/dxos/dxos/packages/core/halo/credentials/src/processor/device-state-machine.ts";
var DeviceStateMachine = class {
  constructor(_params) {
    this._params = _params;
    this.authorizedDeviceKeys = new ComplexMap5(PublicKey7.hash);
    this.deviceChainReady = new Trigger();
  }
  async processCredential(credential) {
    log3("processing credential...", {
      identityKey: this._params.identityKey,
      deviceKey: this._params.deviceKey,
      credential
    }, {
      F: __dxlog_file5,
      L: 35,
      S: this,
      C: (f, a) => f(...a)
    });
    if (isValidAuthorizedDeviceCredential(credential, this._params.identityKey, this._params.deviceKey)) {
      this.deviceCredentialChain = {
        credential
      };
      this.deviceChainReady.wake();
    }
    const assertion = getCredentialAssertion(credential);
    switch (assertion["@type"]) {
      case "dxos.halo.credentials.AuthorizedDevice": {
        invariant4(!this.authorizedDeviceKeys.has(assertion.deviceKey), "Device already added.", {
          F: __dxlog_file5,
          L: 51,
          S: this,
          A: [
            "!this.authorizedDeviceKeys.has(assertion.deviceKey)",
            "'Device already added.'"
          ]
        });
        this.authorizedDeviceKeys.set(assertion.deviceKey, {});
        log3("added device", {
          localDeviceKey: this._params.deviceKey,
          deviceKey: assertion.deviceKey,
          size: this.authorizedDeviceKeys.size
        }, {
          F: __dxlog_file5,
          L: 54,
          S: this,
          C: (f, a) => f(...a)
        });
        this._params.onUpdate?.();
        break;
      }
      case "dxos.halo.credentials.DeviceProfile": {
        invariant4(this.authorizedDeviceKeys.has(credential.subject.id), "Device not found.", {
          F: __dxlog_file5,
          L: 64,
          S: this,
          A: [
            "this.authorizedDeviceKeys.has(credential.subject.id)",
            "'Device not found.'"
          ]
        });
        if (assertion && credential.subject.id.equals(this._params.deviceKey)) {
          log3.trace("dxos.halo.device", {
            deviceKey: credential.subject.id,
            profile: assertion.profile
          }, {
            F: __dxlog_file5,
            L: 67,
            S: this,
            C: (f, a) => f(...a)
          });
        }
        this.authorizedDeviceKeys.set(credential.subject.id, assertion.profile);
        this._params.onUpdate?.();
        break;
      }
    }
  }
};

// packages/core/halo/credentials/src/processor/profile-state-machine.ts
import { log as log4 } from "@dxos/log";
var __dxlog_file6 = "/home/runner/work/dxos/dxos/packages/core/halo/credentials/src/processor/profile-state-machine.ts";
var ProfileStateMachine = class {
  constructor(_params) {
    this._params = _params;
  }
  async processCredential(credential) {
    const assertion = getCredentialAssertion(credential);
    switch (assertion["@type"]) {
      case "dxos.halo.credentials.IdentityProfile": {
        if (!credential.issuer.equals(this._params.identityKey) || !credential.subject.id.equals(this._params.identityKey)) {
          log4.warn("Invalid profile credential", {
            expectedIdentity: this._params.identityKey,
            credential
          }, {
            F: __dxlog_file6,
            L: 34,
            S: this,
            C: (f, a) => f(...a)
          });
          return;
        }
        this.profile = assertion.profile;
        log4("updated profile", {
          identityKey: this._params.identityKey,
          profile: this.profile
        }, {
          F: __dxlog_file6,
          L: 40,
          S: this,
          C: (f, a) => f(...a)
        });
        this._params.onUpdate?.();
        break;
      }
    }
  }
};
export {
  CredentialGenerator,
  DeviceStateMachine,
  FeedStateMachine,
  InvitationStateMachine,
  MemberStateMachine,
  ProfileStateMachine,
  SIGNATURE_TYPE_ED25519,
  SpaceStateMachine,
  canonicalStringify,
  checkCredentialType,
  createAdmissionCredentials,
  createCredential,
  createCredentialMessage,
  createCredentialSignerWithChain,
  createCredentialSignerWithKey,
  createDelegatedSpaceInvitationCredential,
  createDeviceAuthorization,
  credentialTypeFilter,
  defaultSecretProvider,
  defaultSecretValidator,
  generatePasscode,
  getCredentialAssertion,
  getCredentialProofPayload,
  getPresentationProofPayload,
  isValidAuthorizedDeviceCredential,
  signPresentation,
  verifyChain,
  verifyCredential,
  verifyCredentialSignature,
  verifyPresentation,
  verifyPresentationChain,
  verifyPresentationSignature
};
//# sourceMappingURL=index.mjs.map
